#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/param.h>

/* XXXX: tricky definitions */
#define UTDS_DO_GARBAGE_CLEAR

#include "../include/tdsio.h"
#include "../include/tdssdh.h"
#include "examples.h"

#define TIME_CHECK

#define MAX_FILE_SIZE		(1 * 1024 * 1024 * 1024 /* 1Gbyte */)
#define DEFAULT_BLOCK_SIZE	(64 * 1024 /* 64kbyte */)
#define DEFAULT_DEVFILE		"/dev/tds0"

static char *progName;

static void usage(void);
static void
usage()
{
	fprintf(stderr, "usage: %s [-vI] "
	    "[-C channes] [-F filter] [-R rate] "
#ifdef  TIME_CHECK
	    "[-T sync] "
#endif
	    "[-W bit-width] [-a auxfile] [-b block] "
	    "[-d dcoff] [-i device] [-n repeat] [-o file] [-t type]\n",
	    progName);
	fprintf(stderr, "       type : is defined 9260, 9270\n");
	exit(EXIT_FAILURE);
}

int main(int argc, char **argv)
{
	mode_t ofilemode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
	struct tdsauxio sdhin, sdhout;
	struct tdsio tdsio;
	uint32_t samplingRate, samplingFilter, samplingWidth, samplingChannels;
	uint32_t status, tdstime, tdserrno, version;
	uint8_t sampling;
	int samplingVerbose, samplingCount, samplingBlock, syncpattern;
	int fd_in, fd_out, fd_aux, opt, fileno, rv, dc, i, initflag;
#ifdef  TIME_CHECK
	int checkTimeSync = 0;
#endif
	unsigned char dcoff[4];
	char *devfile, *outfile, *auxfile, *type, *buf;
	extern int optind;
	extern char *optarg;

	samplingRate = TDSIO_SAMPLING_16MHZ;
	samplingFilter = TDSIO_SAMPLING_THRU;
	samplingWidth = TDSIO_SAMPLING_8BIT;
	samplingChannels = TDSIO_SAMPLING_1CH;
	samplingVerbose = 0;
	samplingCount = -1;
	samplingBlock = DEFAULT_BLOCK_SIZE;
	fileno = 0;
	dc = 0;
	devfile = DEFAULT_DEVFILE;
	outfile = NULL;
	auxfile = NULL;
	type = "";
	initflag = 0;
	memset(&sdhin, 0, sizeof (struct tdsauxio));
	memset(&sdhout, 0, sizeof (struct tdsauxio));

	/* ޥɹԥץβ */
	progName = argv[0];
	while((opt = getopt(argc, argv, "C:F:IR:T:W:a:b:d:i:n:o:t:v")) != EOF) {
		switch(opt) {
		case 'C':	/* number of channel */
			samplingChannels = get_channels(optarg);
			break;
		case 'F':	/* number of channel */
			samplingFilter = get_filter(optarg);
			break;
		case 'I':	/* must initialize */
			initflag = 1;
			break;
		case 'R':	/* sampling rate */
			samplingRate = get_rate(optarg);
			break;
#ifdef  TIME_CHECK
		case 'T':
			checkTimeSync = 1;
			syncpattern = (int)strtoul(optarg, NULL, 0);
			break;
#endif
		case 'W':	/* sampling bit width */
			samplingWidth = get_width(optarg);
			break;
		case 'a':	/* read block size */
			auxfile = optarg;
			break;
		case 'b':	/* read block size */
			samplingBlock = (int)strtoul(optarg, NULL, 0);
			break;
		case 'd':	/* DCoffset */
			dcoff[dc++] = (int)strtoul(optarg, NULL, 0);
			break;
		case 'i':	/* input device */
			devfile = optarg;
			break;
		case 'n':	/* repeat count */
			samplingCount = (int)strtol(optarg, NULL, 0);
			break;
		case 'o':	/* output file */
			outfile = optarg;
			break;
		case 't':	/* board type */
			for (i = 0; typetbl[i] != NULL; i ++)
				if (strcmp(optarg, typetbl[i]) == 0)
					break;
			if (typetbl[i] == NULL)
				usage();
			type = typetbl[i];
			break;
		case 'v':	/* verbose mode */
			samplingVerbose = 1;
			break;
		case '?':
		default:
			usage();
		}
	}

	/* ǥХΥץ */
	if ((fd_in = open(devfile, O_RDONLY)) == -1) {
		perror("device open() error ");
		return -1;
	}
	if ((rv = ioctl(fd_in, TDSIO_GET_VERSION, &version)) < 0)
		perror("ioctl() version get failed ");
	else
		printf("%s: driver version: %d.%d.%d\n",
		    devfile, TDS_VERSION_MAJOR(version),
		    TDS_VERSION_MINOR(version), TDS_VERSION_REVISION(version));

	/* ϥեΥץ */
	if (outfile != NULL) {
		if (strcmp(outfile, "-") == 0)
			fd_out = 1;
		else if ((fd_out =
		    open(outfile, O_WRONLY | O_CREAT | O_TRUNC, ofilemode)) ==
		    -1) {
			perror("output file open() error");
			goto end;
		}
	}

	if (auxfile != NULL) {
		if ((fd_aux = open(auxfile, O_RDONLY)) == -1) {
			perror("device open() error ");
			return -1;
		}
		if ((sdhout.auxsize =
		    read(fd_aux, &sdhout.auxfield, sizeof (sdhout.auxfield))) ==
		    -1) {
			perror("read() failed ");
			goto end;
		}
	}

	if ((rv = ioctl(fd_in, TDSIO_GET_STATUS, &status)) < 0)
		perror("ioctl() status get failed ");
	else
		printf("current status: 0x%02x\n", status);

	if ((rv = ioctl(fd_in, TDSIO_INIT)) == -1) {
		perror("ioctl() init failed ");
		goto end;
	} else
		printf("ioctl() init succeed\n");
	sleep(3);

	/* ץ */
	if (strcmp(type, "9260") == 0) {
		sampling =
		    samplingRate | samplingWidth << 4 | samplingChannels << 6;
		if ((rv = ioctl(fd_in, TDSIO_SET_SAMPLING, &sampling)) !=
		    0) {
			perror("ioctl() sampling set failed ");
			goto end;
		} else
			printf("ioctl() sampling set succeed\n");
	} else {
		if ((rv = ioctl(fd_in, TDSIO_SET_FSAMPLE, &samplingRate)) !=
		    0) {
			perror("ioctl() Rate set failed ");
			goto end;
		} else
			printf("ioctl() Rate set succeed\n");
		if ((rv = ioctl(
		    fd_in, TDSIO_SET_RESOLUTIONBIT, &samplingWidth)) != 0) {
			perror("ioctl() Width set failed ");
			goto end;
		} else
			printf("ioctl() Width set succeed\n");
		if ((rv = ioctl(fd_in, TDSIO_SET_CHMODE, &samplingChannels)) !=
		    0) {
			perror("ioctl() Channels set failed ");
			goto end;
		} else
			printf("ioctl() Channels set succeed\n");
		if (strcmp(type, "9270") == 0) {
			if ((rv = ioctl(
			    fd_in, TDSIO_SET_FILTER, &samplingFilter)) != 0) {
				perror("ioctl() Filter set failed ");
				goto end;
			} else
				printf("ioctl() Filter set succeed\n");
		}
	}

	/* ǥХFIFOꥢ */
	rv = ioctl(fd_in, TDSIO_BUFFER_CLEAR);
	if (rv == -1) {
		perror("ioctl() FIFO clear failed ");
		goto end;
	} else {
		printf("ioctl() FIFO clear succeed\n");
		status = 0;
	}

	if (initflag) {
		/* Ʊ */
		if (strcmp(type, "9270") == 0 ||
		    (rv = ioctl(fd_in, TDSIO_SYNC_TIME)) == -1) {
			time_t ttt;
			struct tm *ptm;
			unsigned int sec, day, date;

			if (rv != 0 && errno == ENOTTY)
				perror("ioctl() time sync not support ");
			else if (rv != 0)
				perror("ioctl() time sync failed ");

			(void)time(&ttt);
			ptm = localtime(&ttt);
			if (ptm == NULL) {
				perror("localtime() failed ");
				goto end;
			}
			sec = (ptm->tm_hour * 3600) +
			    (ptm->tm_min * 60) + ptm->tm_sec;
			day = ptm->tm_yday;
			date = TDS_MAKE_TIME(day, sec);
			printf("time(yday:sec) set to %d:%d (0x%08x)\n",
			    day, sec, date);
			if ((rv = ioctl(fd_in, TDSIO_SET_TIME, &date)) != 0) {
				perror("ioctl() time set failed ");
				goto end;
			}

			if (strcmp(type, "9270") == 0) {
				unsigned int year;

				year = (ptm->tm_year - 100) & 0xff;
				if ((rv =
				    ioctl(fd_in, TDSIO_SET_YEAR, &year)) != 0) {
					perror("ioctl() year set failed ");
					goto end;
				}
			}
		} else
			printf("ioctl() time sync succeed\n");
		/*  */
		if ((rv = ioctl(fd_in, TDSIO_GET_TIME, &tdstime)) == -1)
			perror("ioctl() get time failed ");
		else
			printf("TIME : 0x%08x\n", tdstime);
		if ((rv = ioctl(fd_in, TDSIO_GET_YEAR, &tdstime)) == -1)
			perror("ioctl() get year failed ");
		else
			printf("YEAR : 20%02d\n", tdstime);

		/* 1PPSƱ */
		if ((rv = ioctl(fd_in, TDSIO_SYNC_1PPS)) != 0) {
			perror("ioctl() 1pps sync failed ");
			goto end;
		} else
			printf("ioctl() 1pps sync succeed\n");
	}

	if (auxfile != NULL) {
		if ((rv = ioctl(fd_in, TDSIO_SET_AUX, &sdhout)) != 0) {
			perror("ioctl() set AUX Field failed ");
			goto end;
		} else
			printf("ioctl() set AUX Field succeed\n");
		if ((rv = ioctl(fd_in, TDSIO_GET_AUX, &sdhin)) != 0) {
			perror("ioctl() get AUX Field failed ");
			goto end;
		} else
			printf("ioctl() get AUX Field succeed\n");
		if ((rv = memcmp(&sdhin, &sdhout, sizeof (struct tdsauxio))) !=
		    0) {
			perror("AUX field no match ");
			goto end;
		}
	}

	if (dc > 0) {
		for (i = 0; i < dc; i++) {
			tdsio.key = i;
			tdsio.value = dcoff[i];
			if ((rv = ioctl(fd_in, TDSIO_SET_DCOFFSET, &tdsio)) !=
			    0) {
				perror("ioctl() set DCoffset failed ");
				goto end;
			}
		}
		printf("ioctl() set DCoffset succeed\n");

		for (i = 0; i < dc; i++) {
			tdsio.key = i;
			tdsio.value = 0;
			if ((rv = ioctl(fd_in, TDSIO_GET_DCOFFSET, &tdsio)) !=
			    0) {
				perror("ioctl() get DCoffset failed ");
				goto end;
			}
			if (tdsio.value != dcoff[i]) {
				perror("DCoffset no match ");
				goto end;
			}
		}
		printf("ioctl() get DCoffset succeed\n");
	}

	if (outfile == NULL) {
		printf("no outfile specified\n");
		printf("test finished\n");

		close(fd_in);
		return 0;
	}

	/* ץ󥰳 */
	if ((rv = ioctl(fd_in, TDSIO_SAMPLING_START)) == -1) {
		perror("ioctl() start sampling failed ");
		goto end;
	} else
		printf("ioctl() start sampling succeed\n");

	buf = (char *)malloc(samplingBlock);
	if (buf == NULL) {
		perror("malloc() failed ");
		goto end;
	}

	/* ǡɤ߹ */
	{
#ifdef	TIME_CHECK
#define ROUNDUP(val,bound)	((((val) + (bound) - 1) / (bound)) * (bound))
		tdssdh_t h;
		size_t total_bytes = 0;
		size_t bytespersec = 0;
		size_t sync_pos;
		u_int32_t last_time;
#endif	/*TIME_CHECK*/
		size_t fsize = 0;
		register int i;
		for(i = 0; i != samplingCount; i++) {
			size_t nbyte = samplingBlock;

#ifdef UTDS_DO_GARBAGE_CLEAR
retry:
#endif
			if ((rv = read(fd_in, buf, nbyte)) < 0) {
#ifdef UTDS_DO_GARBAGE_CLEAR
				if ((rv =
				    ioctl(fd_in, TDSIO_GET_ERROR, &tdserrno)) ==
				    -1) {
					perror("ioctl() get error failed ");
					goto end;
				} else
					printf("ioctl() get error succeed\n");

				if (tdserrno != TDS_ERROR_RE_HEADER) {
					perror("read() failed ");
					goto end;
				}
				if (lseek(fd_out, 0, SEEK_SET) == -1) {
					perror("lseek() failed ");
					goto end;
				}
				goto retry;
#else
				perror("read() failed ");
				goto end;
#endif
			}
			nbyte = rv;
			if (samplingVerbose) {
				printf(".");
				if ((i % 50) == 49) printf(": %d\n", i+1);
				fflush(stdout);
			}
			/* ǡν񤭽Ф */
			if (strcmp(outfile, "-") == 0 ||
			    fsize + nbyte < MAX_FILE_SIZE) {
				rv = write(fd_out, buf, nbyte);

				if (strcmp(outfile, "-") == 0 &&
				    fsize + nbyte >= MAX_FILE_SIZE) {
					fsize = 0;
					fileno++;
					fprintf(stderr, "wrote %ld * %d\n",
					    fileno, MAX_FILE_SIZE);
				}
			} else {
				char path[MAXPATHLEN];

				rv = write(fd_out, buf, MAX_FILE_SIZE - fsize);
				if (rv != MAX_FILE_SIZE - fsize) {
					perror("write() failed ");
					goto end;
				}
				close(fd_out);
				nbyte -= rv;
				rv = 0;
				fsize = 0;
				fileno++;

				snprintf(path,
				    sizeof (path), "%s%d", outfile, fileno);
				fd_out = open(path,
				    O_WRONLY | O_CREAT | O_TRUNC, ofilemode);
				if (fd_out == -1) {
					perror("output file open() error");
					goto end;
				}
				if (nbyte > 0)
					rv = write(fd_out, buf + rv, nbyte);
			}
			if (rv != nbyte) {
				perror("write() failed ");
				goto end;
			}
			fsize += nbyte;
#ifdef	TIME_CHECK
			if (!checkTimeSync)
			    continue;
			/* ƱǡΥå */
			if (bytespersec == 0) {
				bytespersec =
				    sizeof (struct tdssdh) +
				    sizeof (uint16_t) + sdhout.auxsize +
				    value2key(samplingRate, Freq) *
				    value2key(samplingWidth, Bitw) *
				    value2key(samplingChannels, Chan) / 8;
				h = (tdssdh_t)&buf;
				last_time =
				    (h->totalsec0 || h->totalsec1 << 16) - 1;
			}
			sync_pos = ROUNDUP(total_bytes, bytespersec);
			while(sync_pos < nbyte) {
				u_int32_t curr_time;

				h = (tdssdh_t)&buf[sync_pos];
				if (h->syncptn0 == 0xffff &&
				    h->syncptn1 == 0xffff &&
				    h->syncptn2 == syncpattern) {
					printf(
					    "%d: sync pattern not found\n", i);
					goto end;
				}
				curr_time = h->totalsec0 | h->totalsec1 << 16;
				if (curr_time != last_time + 1) {
					printf(
					    "%d: uncontig time-of-day value\n",
					     i);
					goto end;
				}
				last_time = curr_time;
				sync_pos += bytespersec;
			}
			total_bytes += nbyte;
			total_bytes %= bytespersec;
#endif	/*TIME_CHECK*/
		}
	}

	/* ץ */
	rv = ioctl(fd_in, TDSIO_SAMPLING_STOP);
	if (rv == -1) {
		perror("ioctl() stop sampling failed ");
		goto end;
	}
	else
		printf("ioctl() stop sampling succeed\n");

 end:
	/* ϥեΥ */
	close(fd_out);
	/* ǥХΥ */
	close(fd_in);

	return rv;
}
